@@ -7,6 +7,10 @@ import re
 import sys
 import platform
 import subprocess
+import site
+import webbrowser
+from datetime import datetime
+from dsframework.documentation.scripts import doxygen_handler
 
 isWindows = platform.system() == 'Windows'
 
@@ -33,8 +37,17 @@ directories = {
     'dsp': '',
 }
 
+##
+# @file
+# @brief This file defines and implements the framework CLI. It helps us to:\n
+#        Create: Project/Forcer/Predictable/Tester-files/Deploy-files/Cloud-Eval-Files  \n
+#        Run   : Server / Evaluation (CSV,Cloud) \n
 
 class AliasedGroup(click.Group):
+    """! AliasedGroup implements click.Group base class."""
+
+    ##
+    # @hidecallgraph @hidecallergraph
     def get_command(self, ctx, cmd_name):
         try:
             cmd_name = ALIASES[cmd_name].name
@@ -88,6 +101,17 @@ def cli():
 
     dsf-cli evaluate input-csv-file batch-size
 
+    Generate project documentation:
+
+    dsf-cli generate documentation full
+
+    Generate documentation - without base classes:
+
+    dsf-cli generate documentation simple
+
+    dsf-cli show-documentation
+
+    dsf-cli upload-documentation
     """
 # @click.option("--type", prompt="type", help="type of component")
 # @click.option("--project_name", prompt="project_name", help="project_name")
@@ -109,6 +133,7 @@ def generate(gen_type, project_name):
 
 @cli.command()
 def run_server():
+    """Using /server/main.py """
     currentPipelineFolder = os.path.abspath(os.getcwd())
     currentParentFolder = os.path.join(currentPipelineFolder,"..")
     os.environ["PYTHONPATH"] = currentPipelineFolder
@@ -119,6 +144,7 @@ def run_server():
 @click.argument('tag')
 @click.argument('env', default="prd")
 def cloud_eval(tag, env):
+#    """Cloud evaluation running on AWS """
     currentPipelineFolder = os.path.abspath(os.getcwd())
     currentParentFolder = os.path.join(currentPipelineFolder,"..")
     os.environ["PYTHONPATH"] = currentPipelineFolder
@@ -162,17 +188,21 @@ def cloud_eval(tag, env):
 
 @cli.command()
 def create_deploy_files():
+    """Create deploy files (created through 'Create project', but we can use it as standalone function """
     copy_deploy_files('')
 
 
 @cli.command()
 def create_cloud_eval_files():
+    """Create cloud evaluation files (created through 'Create project', but we can use it as standalone function """
     currentPipelineFolder = os.path.abspath(os.getcwd())
     copy_cloud_eval_files(currentPipelineFolder)
 
 
 @cli.command()
 def create_tester_files():
+    """!Create tester files (created through 'Create project', but there is an option
+       We can use it as standalone function """
     create_tester_env()
 
 
@@ -180,6 +210,15 @@ def create_tester_files():
 @click.argument('csv_file_path')
 @click.argument('batch_size')
 def evaluate(csv_file_path, batch_size):
+    """! Quick 'Initial test' to check the module.
+
+    @verbatim
+    Args:
+        csv_file_path: CSV file path.
+        batch_size: The size that will be taken every iteration from the CSV - in order to feed the model.
+    @endverbatim
+    """
+
     if os.path.isfile(csv_file_path):
         current_pipeline_folder = os.path.abspath(os.getcwd())
         os.environ["PYTHONPATH"] = current_pipeline_folder
@@ -189,33 +228,226 @@ def evaluate(csv_file_path, batch_size):
         click.echo('file: ' + csv_file_path + ' not found')
 
 
+@cli.command()
+def show_documentation():
+    launch_doc_webbrowser()
+
+
+def launch_doc_webbrowser():
+    current_pipeline_folder = os.path.abspath(os.getcwd())
+
+    current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
+    if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
+        current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")
+
+    project_html_path = os.path.join(current_doc_folder, 'project_docs', 'html', 'index.html')
+    if os.path.isfile(project_html_path):
+        webbrowser.open('file://' + project_html_path)
+    else:
+        click.echo('Project documentation not found, please generate documentation using: \'dsf-cli generate '
+                   'documentation full\'')
+
+
 ALIASES = {
     "g": generate
 }
 
 
 def generate_project(project_name):
+    """! Generating new project (with the relevant project_name)\n
+    How to generate project :
+    > dsf-cli g project my-new-project
+    @verbatim
+    Args:
+            project_name: Project name will be used in the creation of the files.
+    @endverbatim
+    """
     # project_name = clean_name(project_name)
     click.echo('Generating project: ' + project_name)
     create_project(project_name)
 
 
 def generate_forcer(fileName):
+    """! Generate new forcer - can help us, if we need forcer on an existing project """
     fileName = clean_name(fileName)
     create_exist_pipeline_file('forcer', fileName)
 
 
 def generate_predictable(fileName):
+    """! Generate new predictable - can help us, if we need predictable on an existing project """
     fileName = clean_name(fileName)
     create_exist_pipeline_file('predictable', fileName)
 
 
+def generate_documentation(simple="full"):
+    """! Document project CLI
+
+        Usage:
+            dsf-cli generate documentation full
+    """
+    if simple == 'full':
+        click.echo('Documenting project .....')
+    else:
+        click.echo('Documenting project (without base).....')
+
+    # Get folders
+    current_pipeline_folder = os.path.abspath(os.getcwd())
+    current_doc_folder = os.path.join(current_pipeline_folder, "documentation")
+    if os.path.isdir(os.path.join(current_pipeline_folder, "dsframework", "documentation")):
+        current_doc_folder = os.path.join(current_pipeline_folder, "dsframework", "documentation")
+
+    # Validity checks
+    if not os.path.isfile(os.path.join(current_doc_folder, 'project_doxyfile')):
+        click.echo("Documentation folder not found, please change directory to project root folder.")
+        return
+
+    venv_path = os.getenv('VIRTUAL_ENV')
+
+    try:
+        if simple == 'full':
+            os.environ["EXCLUDE_PATH"] = get_exclude_string(venv_path)
+        else:
+            os.environ["EXCLUDE_PATH"] = venv_path
+    except Exception as e:
+        click.echo(f'Switching to simple: {e}')
+        os.environ["EXCLUDE_PATH"] = venv_path
+
+    click.echo(f'Creating project documentation in {os.path.join(current_doc_folder, "project_docs")}')
+
+    try:
+        # Run doxygen as subprocess
+        subprocess.call('doxygen project_doxyfile', shell=True, cwd=current_doc_folder)
+    except Exception as e:
+        click.echo(f'Project documentation failed!, exception: {e}')
+    else:
+        click.echo('Project documentation done!')
+        launch_doc_webbrowser()
+
+
+def get_exclude_string(venv_path):
+    packages_path = site.getsitepackages()  # Get packages folder.
+
+    # Get list of packages to exclude, except 'dsframework'
+    pkg_exclude_folders = get_excluded_path_list('(?!.*dsframework)', packages_path[0], 'folders')
+
+    # Get list of scattered file in packages folder
+    pkg_exclude_files = get_excluded_path_list('(?!.*api_cli.py)', packages_path[0], 'files')
+
+    # Get a list of files and folders in dsframework, except 'base' folder.
+    pkg_exclude_dsf = get_excluded_path_list('(?!.*base)', os.path.join(packages_path[0], 'dsframework'), 'both')
+
+    # Get venv all the rest the folders to exclude, except 'venv/lib'
+    exclude_venv = get_excluded_path_list('(?!.*lib)', venv_path, 'both')
+
+    # Combine two lists to one string: exclude_folders and exclude_folders
+    exclude_full_list = pkg_exclude_folders + exclude_venv + pkg_exclude_files + pkg_exclude_dsf
+    exclude_string = " \\ ".join(map(str, exclude_full_list))  # Seperated by backslash
+
+    return exclude_string
+
+
+def get_excluded_path_list(regex_str, path, get_type):
+
+    if get_type == 'folders':
+        path_list = [os.path.join(path, name) for name in os.listdir(path)
+                     if os.path.isdir(os.path.join(path, name))]
+    elif get_type == 'files':
+        path_list = [os.path.join(path, name) for name in os.listdir(path)
+                     if os.path.isfile(os.path.join(path, name))]
+    else:  # 'both'
+        path_list = [os.path.join(path, name) for name in os.listdir(path)]
+
+    if regex_str != '':
+        regex = re.compile(regex_str)
+        return list(filter(regex.match, path_list))
+    else:
+        return path_list
+
+
+@cli.command()
+def upload_documentation():
+    """! Upload project documentation CLI
+
+        Usage:
+            dsf-cli upload-documentation
+    """
+
+    # Set folders
+    current_working_path = os.getcwd()
+    project_name = os.path.split(current_working_path)[1]
+    current_pipeline_folder = os.path.abspath(os.getcwd())
+    current_doc_folder = os.path.join(current_pipeline_folder, 'documentation', 'project_docs')
+    html_folder_name = 'html'
+
+    google_gs_path = 'gs://dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'
+    project_gs_folder = f'{google_gs_path}{project_name}'
+    google_storage_path = 'https://storage.cloud.google.com/dozi-stg-ds-apps-1-ds-apps-ds-portal/documentation/'
+    base_tag_path = google_storage_path + project_name + '/' + html_folder_name + '/'
+
+    # Validity checks
+    if not os.path.isfile(os.path.join(current_doc_folder, html_folder_name, 'index.html')):
+        click.echo('Documentation folder not found, please change directory to project root folder.')
+        return
+
+    click.echo(f'Uploading {project_name} documentation started....')
+
+    # Create backup of current docs on google storage
+    click.echo(f'Creating backup of {project_name} current documentation....')
+
+    backup_project_name = datetime.today().strftime('%Y_%m_%d_%H_%M_%S') + '_' + project_name
+    backup_destination_folder = f'{google_gs_path}backups/{backup_project_name}'
+
+    call_gsutil_command(f'gsutil -m mv {project_gs_folder} {backup_destination_folder}',
+                        project_name, current_doc_folder, 'Backup current documentation')
+
+    # Modify project documentation, add base tag and paths
+    click.echo(f'Adding <base> tag to {project_name} documentation....')
+    try:
+        html_list = doxygen_handler.find_files(os.path.join(current_doc_folder, html_folder_name), 'html')
+        css_list = doxygen_handler.find_files(os.path.join(current_doc_folder, html_folder_name), 'css')
+
+        doxygen_handler.modify_files(html_list, '<head>', '<head><base href=\"url_tag\">', base_tag_path)
+        doxygen_handler.modify_files(css_list, "url(\'", "url(\'url_tag", base_tag_path)
+        doxygen_handler.modify_files(css_list, "url(\"", "url(\"url_tag", base_tag_path)
+    except Exception as e:
+        click.echo(f'Adding <base> tag failed with exception: {e}')
+        return
+    else:
+        click.echo('Adding <base> tag done!')
+
+    # Upload current documentation
+    call_gsutil_command(f'gsutil -m cp -r {os.path.join(current_doc_folder, html_folder_name)} {project_gs_folder}',
+                        project_name, current_doc_folder, 'Uploading documentation')
+
+    click.echo(f'Uploading {project_name} documentation done!!!')
+
+
+def call_gsutil_command(gsutil_command, project_name, working_folder, command_desc):
+    click.echo(f'{command_desc} {project_name} ....')
+
+    try:
+        subprocess.call(gsutil_command, shell=True, cwd=working_folder)
+    except Exception as e:
+        click.echo(f'{command_desc} failed with exception: {e}')
+        return
+    else:
+        click.echo(f'{command_desc} done!')
+
+
 def clean_name(name):
+    """! Change the name from 'word1-word2' --> 'word1_word2' """
     name = name.replace('-', '_')
     return name
 
 
 def create_folders(project_name):
+    """! Create all needed folders for the new project creation
+    @verbatim
+    Args:
+        project_name: Project name
+    @endverbatim
+    """
+
     global directories
     directories['main'] = project_name
     if not os.path.exists(directories['main']):
@@ -238,10 +470,16 @@ def create_folders(project_name):
     create_main_folders('dsp', 'tester')
     create_main_folders('trainer', 'main')
     create_main_folders('server', 'main')
-    create_main_folders('documentation', 'main')
 
 
 def create_project(project_name):
+    """! Create project from templates.
+    @verbatim
+    Args:
+        project_name: Project name
+    @endverbatim
+    """
+
     create_folders(project_name)
     original_project_name = project_name
     project_name = clean_name(project_name)
@@ -283,7 +521,8 @@ def create_project(project_name):
     create_server_config_json()
     copy_deploy_files(directories['main'])
     copy_cloud_eval_files(directories['main'])
-    copy_documentation_files(directories['documentation'])
+    copy_documentation_files(os.path.join(directories['main'], 'documentation'))
+    update_doc_configuration(project_name, os.path.join(directories['main'], 'documentation'))
 
     change_to_project_dir()
     run_dvc_init()
@@ -291,6 +530,14 @@ def create_project(project_name):
 
 
 def create_main_folders(targetDir, baseDir):
+    """! Create folder if not already exists
+    @verbatim
+    Args:
+        targetDir: the new folder to create
+        baseDir  : where we want to create the new folder.
+    @endverbatim
+    """
+
     global directories
     directories[targetDir] = directories[baseDir] + '/' + targetDir
     if not os.path.exists(directories[targetDir]):
@@ -298,31 +545,85 @@ def create_main_folders(targetDir, baseDir):
 
 
 def create_pipeline_file(project_name, folder, pipelineType, createInitFile=True):
+    """! Takes the template file and replace it with new one with project_name
+    @verbatim
+    Args:
+        project_name : The project name
+        folder       : The folder we want to create the new file.
+        pipelineType : File type in the pipeline.
+        createInitFile(=True): Create init file - if doesnt exist.
+    @endverbatim
+    """
+
     data = read_template_file(pipelineType)
     replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)
 
 
 def create_schema_file(project_name, folder, pipelineType, createInitFile=True):
+    """! Create schema files (read the template and replace it )
+    @verbatim
+    Args:
+        project_name : The project name
+        folder       : The folder we want to create the new file.
+        pipelineType : File type. TODO change pipelineType name
+        createInitFile(=True): Create init file - if doesnt exist.
+    @endverbatim
+    """
     data = read_template_file('schema/' + pipelineType)
     replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)
 
 
 def create_server_file(project_name, folder, pipelineType, createInitFile=True):
+    """! Create server file (read the template and use him to create the new file)
+    @verbatim
+    Args:
+        project_name : The project name
+        folder       : The folder we want to create the new file.
+        pipelineType : File type. TODO change pipelineType name
+        createInitFile(=True): Create init file - if doesnt exist.
+    @endverbatim
+    """
     data = read_template_file('tester/server/' + pipelineType)
     replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)
 
 
 def create_tester_file(project_name, folder, test_file_type, createInitFile=True):
+    """! Create tester file (read the template and use him to create the new file)
+    @verbatim
+    Args:
+        project_name   : The project name.
+        folder         : The folder we want to create the new file.
+        test_file_type : File type.
+        createInitFile(=True): Create init file - if doesnt exist.
+    @endverbatim
+    """
     data = read_template_file('tester/' + test_file_type)
     replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)
 
 
 def create_dsp_file(project_name, folder, test_file_type, createInitFile=True):
+    """! Create DSP file (read the template and use him to create the new file)
+    @verbatim
+        Args:
+            project_name   : The project name.
+            folder         : The folder we want to create the new file.
+            test_file_type : File type.
+            createInitFile(=True): Create init file - if doesnt exist.
+    @endverbatim
+    """
     data = read_template_file('tester/dsp/' + test_file_type)
     replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)
 
 
 def create_exist_pipeline_file(type, fileName):
+    """! Create a new component in the pipeline and integrate it (for example forcer or predictable)
+    @verbatim
+    Args:
+        type      : New component type (forcer or predictable)
+        fileName  : New file name.
+    @endverbatim
+    """
+
     pipelineType = type + 's'
     folder = 'pipeline/' + pipelineType
     if os.path.exists(folder):
@@ -351,6 +652,7 @@ def create_tester_env():
     This function will create a testing environment, if one does not already exist in the project.
     It will create the testing, evaluation, and relevant schema files.
     """
+
     project_name = os.getcwd().split('/')[-1]
     dst_folder = 'tester/'
     schema_dst_folder = 'tester/test_schema/'
@@ -391,6 +693,7 @@ def create_tester_env():
 
 
 def read_template_file(filename):
+    """! Read \<filename\>_template.py  from the cli dir """
     with open(os.path.join(__location__, 'dsframework/cli/' + filename + '_template.py'), 'r') as file:
         data = file.read()
         return data
@@ -409,6 +712,17 @@ def write_to_file(file_path, data):
 
 
 def replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile):
+    """! Get & Replace the template file and create new one with project_name
+    @verbatim
+    Args:
+        project_name : The project name
+        folder       : The folder we want to create the new file.
+        pipelineType : Which File.
+        data         : The template file.
+        createInitFile(=True): Create init file - if doesnt exist.
+    @endverbatim
+    """
+
     pipeline_type_no_underscore = to_capitalize_no_underscore(pipelineType)
     project_name_no_underscore = to_capitalize_no_underscore(project_name)
     className = project_name_no_underscore + pipeline_type_no_underscore
@@ -440,6 +754,14 @@ def create_predictable_file(fileName):
 
 
 def create_file(new_file_path, data):
+    """! Create new file and write the data file into it
+    @verbatim
+    Args:
+        new_file_path : New file name path.
+        data          : The content of the new file.
+    @endverbatim
+    """
+
     if not os.path.exists(new_file_path):
         f = open(new_file_path, "w")
         f.write(data)
@@ -447,6 +769,16 @@ def create_file(new_file_path, data):
 
 
 def create_init_file(init_path, init_export):
+    """! if init file doesnt exist - create it and add the export line
+         if init file exist - check if the init_export line appears in the init file : if not add it.
+
+    @verbatim
+    Args:
+        init_path   : Init file.
+        init_export : Import line to add if needed.
+    @endverbatim
+    """
+
     if not os.path.exists(init_path):
         f = open(init_path, "w")
         f.write(init_export)
@@ -461,7 +793,23 @@ def create_init_file(init_path, init_export):
             f.close()
 
 
+def save_file(file_name_path, data):
+    f = open(file_name_path, "w")
+    f.write(data)
+    f.close()
+
+
 def inject_to_pipeline(fileName, type, className, new_init_export):
+    """! Building the component in the pipeline file
+    @verbatim
+    Args:
+        fileName       : New file name.
+        type           : New component type (forcer or predictable)
+        className      : New component class.
+        new_init_export: Create init file - if doesnt exist.
+    @endverbatim
+    """
+
     file_path = 'pipeline/pipeline.py'
     if os.path.exists(file_path):
         data = read_file(file_path)
@@ -547,6 +895,7 @@ def inject_to_pipeline(fileName, type, className, new_init_export):
 
 
 def create_project_config_yaml():
+    """! Creates /pipeline/config.yaml from dsframework/cli/config.yaml """
     with open(os.path.join(__location__, 'dsframework/cli/config.yaml'), 'r') as file:
         data = file.read()
         data = data.replace('generatedDirectory', directories['main'])
@@ -558,6 +907,7 @@ def create_project_config_yaml():
 
 
 def create_project_config_json():
+    """! Creates /config/config.json from dsframework/cli/config.json """
     with open(os.path.join(__location__, 'dsframework/cli/config.json'), 'r') as file:
         data = file.read()
         data = data.replace('generatedDirectory', directories['main'])
@@ -569,6 +919,9 @@ def create_project_config_json():
 
 
 def create_project_gitignore():
+    """! Creates /.gitignore from dsframework/cli/.gitignore
+         Which means all files that appear in the project directory and should not be uploaded to the git repository.
+    """
     with open(os.path.join(__location__, 'dsframework/cli/.gitignore'), 'r') as file:
         data = file.read()
         new_file = directories['main'] + '/.gitignore'
@@ -579,6 +932,7 @@ def create_project_gitignore():
 
 
 def create_server_config_json():
+    """! Creates server/cors_allowed_origins from dsframework/cli/cors_allowed_origins   """
     with open(os.path.join(__location__, 'dsframework/cli/cors_allowed_origins.json'), 'r') as file:
         data = file.read()
         data = data.replace('generatedDirectory', directories['main'])
@@ -590,10 +944,18 @@ def create_server_config_json():
 
 
 def change_to_project_dir():
+    """! Change the OS dir path """
     os.chdir(directories['main'])
 
 
 def run_dvc_init():
+    """! Running dvc init and doing the following:
+    1. Defines where all the dvc will be uploaded (DVC_BUCKET=)
+    2. dvc init    - generate a local repository
+    3. dvc config
+    4. dvc remote add - connect my local repo -> remote
+    5. dvc remote modify
+    """
     dir_path = os.getcwd()
     if not os.path.isdir(dir_path + '/.dvc'):
         command = 'dvc_init.sh'
@@ -603,6 +965,9 @@ def run_dvc_init():
 
 
 def run_git_init():
+    """! Using git_init.sh to create a new project on git
+        https://git.zoominfo.com/dozi/{name-your-service}
+    """
     dir_path = os.getcwd()
     if not os.path.isdir(dir_path + '/.git'):
         command = 'git_init.sh'
@@ -612,6 +977,7 @@ def run_git_init():
 
 
 def copy_deploy_files(main_dir):
+    """! Copy deploy files (from dsframework/cli/tester/deploy_files """
     currentPipelineFolder = os.path.basename(os.getcwd())
     if main_dir:
         currentPipelineFolder = main_dir
@@ -650,6 +1016,7 @@ def copy_deploy_files(main_dir):
 
 
 def copy_cloud_eval_files(main_dir):
+    """! Copy cloud evaluation files """
     currentPipelineFolder = os.path.basename(os.getcwd())
     if main_dir:
         currentPipelineFolder = main_dir
@@ -709,16 +1076,32 @@ def copy_cloud_eval_files(main_dir):
 
 
 def copy_documentation_files(target_dir):
+    """! Copy documentation folder to newly created project.
+
+    Args:
+        target_dir: Newly created project location
+    """
     source_dir = os.path.join(__location__, 'dsframework/documentation/')
+    shutil.copytree(source_dir, target_dir, ignore=shutil.ignore_patterns('dsf*'))
 
-    for (dirpath, dirname, filenames) in os.walk(source_dir):
-        doc_files = [os.path.join(dirpath, file) for file in filenames if file != 'dsf_doxyfile']
 
-        for file_path in doc_files:
-            shutil.copy2(file_path, target_dir)
+def update_doc_configuration(project_name, target_dir):
+    """! Update project doxygen configuration file, adding project name.
+
+    Args:
+        project_name: Newly created project name
+        target_dir: Configuration file location.
+    """
+    project_name_no_underscore = to_capitalize_no_underscore(project_name)
+    source_dir = os.path.join(target_dir, 'project_doxyfile')
+    data = read_file(source_dir)
+    data = data.replace('generatedProjectName', project_name_no_underscore)
+    save_file(source_dir, data)
 
 
 def to_capitalize_no_underscore(text):
+    """! Capitalize each word with no underscore
+     for example : Moving from test_name -> TestName"""
     return ''.join(elem.capitalize() for elem in text.split('_'))
 
 
@@ -728,6 +1111,12 @@ def clean_existing_project_name(text):
 
 
 def let_user_pick(options):
+    """! Let the user pick a number from couple of options
+    @verbatim
+    Args:
+        options  : couple of options for user decision.
+    @endverbatim
+    """
     print("Please choose:")
     for idx, element in enumerate(options):
         print("{}) {}".format(idx+1, element))
